package org.gcube.application.cms.plugins.implementations;

import com.sun.xml.internal.xsom.impl.scd.Step;
import lombok.extern.slf4j.Slf4j;
import org.bson.Document;
import org.gcube.application.cms.plugins.IndexerPluginInterface;
import org.gcube.application.cms.plugins.MaterializationPlugin;
import org.gcube.application.cms.plugins.implementations.AbstractLifeCycleManager;
import org.gcube.application.cms.plugins.LifecycleManager;
import org.gcube.application.cms.plugins.faults.*;
import org.gcube.application.cms.plugins.model.PluginDescriptor;
import org.gcube.application.cms.plugins.model.StepAccess;
import org.gcube.application.cms.plugins.reports.*;
import org.gcube.application.cms.plugins.requests.*;
import org.gcube.application.cms.plugins.model.ComparableVersion;
import org.gcube.application.cms.serialization.Serialization;
import org.gcube.application.geoportal.common.model.configuration.Configuration;
import org.gcube.application.geoportal.common.model.configuration.Index;
import org.gcube.application.geoportal.common.model.document.accounting.User;
import org.gcube.application.geoportal.common.model.document.lifecycle.LifecycleInformation;
import org.gcube.application.geoportal.common.model.document.lifecycle.TriggeredEvents;
import org.gcube.application.geoportal.common.model.rest.ConfigurationException;
import org.gcube.application.geoportal.common.model.useCaseDescriptor.HandlerDeclaration;
import org.gcube.application.geoportal.common.utils.Files;

import java.util.ArrayList;
import java.util.List;

@Slf4j
public class Default3PhaseManager extends AbstractLifeCycleManager implements LifecycleManager {

    private static class Phases {
        public static final String PENDING_APPROVAL="Pending Approval";
        public static final String PUBLISHED="PUBLISHED";

    }


    private static class STEPS{
        public static final String SUBMIT="SUBMIT-FOR-REVIEW";
        public static final String REJECT="REJECT-DRAFT";
        public static final String APPROVE="APPROVE-SUBMITTED";
    }

    private static class PARAMETERS{
        public static final String NOTES="notes";
    }


    protected PluginDescriptor DESCRIPTOR=new PluginDescriptor("DEFAULT-3PHASE", PluginDescriptor.BaseTypes.LIFECYCLE_MANAGER);

    public Default3PhaseManager() {
        DESCRIPTOR.setDescription("Default 3-phase lifecycle manager. This plugin supports a simple moderated publication lifecycle.");
        DESCRIPTOR.setVersion(new ComparableVersion("1.0.0"));
    }

    protected static boolean canInvokeStep(String stepID, User u, HandlerDeclaration config) throws ConfigurationException {
        List l =config.getConfiguration().get("step_access", List.class);
        if(l==null|| l.isEmpty()) throw new ConfigurationException("Missing Role management in UCD");
        boolean existingRule=false;
        for (Object o : l) {
            StepAccess a= Serialization.convert(o, StepAccess.class);
            if(a.getStepId().equals(stepID)) {
                existingRule = true;
                log.trace("Found role {}",a);
                for (String role : a.getRoles())
                    if (u.getRoles().contains(role))
                        return true;
            }
        }
        if(!existingRule) throw new ConfigurationException("Missing step "+stepID+" access definition");
        // nothing matches
        return false;
    }


    @Override
    public Configuration getCurrentConfiguration(BaseRequest req) throws ConfigurationException {
        Configuration toReturn = super.getCurrentConfiguration(req);
        toReturn.setIndexes(new ArrayList<>());

        IndexerPluginInterface indexerPlugin;
        indexerPlugin = (IndexerPluginInterface) pluginManager.getById("SDI-Indexer-Plugin");
        BaseRequest indexRequest = new BaseRequest(req.getUseCaseDescriptor(),req.getCaller(),req.getContext());

        // Info on Public index
        try {
            indexRequest.setCallParameters(getPublicIndexParams(req));
            Index publicIndex = indexerPlugin.getIndex(indexRequest);
            publicIndex.put("flag", "public");
            toReturn.getIndexes().add(publicIndex);
        }catch(ConfigurationException e){
            toReturn.addErrorMessage("Unable to gather information on public GIS Centroids Index : "+e.getMessage());
            log.error("Unable to gather information on public GIS Centroids Index",e);
        }
        // Info on internal_index
        try {
            indexRequest.setCallParameters(getInternalIndexParams(req));
            Index internalIndex = indexerPlugin.getIndex(indexRequest);
            internalIndex.put("flag", "internal");
            toReturn.getIndexes().add(internalIndex);
        }catch(ConfigurationException e){
            toReturn.addErrorMessage("Unable to gather information on internal GIS Centroids Index : "+e.getMessage());
            log.error("Unable to gather information on internal GIS Centroids Index",e);
        }
        return toReturn;
    }


    private Document getInternalIndexParams(BaseRequest req){
        Document callParameters = new Document();

        callParameters.put("workspace", Files.fixFilename(req.getUseCaseDescriptor().getId()+"_internal_"+req.getContext().getName()));
        callParameters.put("indexName",Files.fixFilename(req.getUseCaseDescriptor().getId()+"_internal_"+req.getContext().getName()+"_centroids"));
        return callParameters;
    }

    private Document getPublicIndexParams(BaseRequest req){
        Document callParameters = new Document();
        callParameters.put("workspace",Files.fixFilename(req.getUseCaseDescriptor().getId()+req.getContext().getName()));
        callParameters.put("indexName",Files.fixFilename(req.getUseCaseDescriptor().getId()+req.getContext().getName()+"_centroids"));
        return callParameters;
    }


    @Override
    public EventExecutionReport onEvent(EventExecutionRequest request) throws EventException, InvalidPluginRequestException {
        EventExecutionReport report=super.onEvent(request);
        TriggeredEvents info = report.getToSetLifecycleInformation().getLastEvent();
        try {
            switch(request.getEvent()){
                case EventExecutionRequest.Events.ON_INIT_DOCUMENT:
                    // Set Defaults as for on update
                case EventExecutionRequest.Events.ON_UPDATE_DOCUMENT: {
                    log.debug("Setting default values..");
                    report=setDefault(report);
                    break;
                }
                case EventExecutionRequest.Events.ON_DELETE_DOCUMENT: {
                    //DELETE ALL
                    break;
                }
                default: throw new EventException("Unexpected Event "+request.getEvent());
            }
        }catch (EventException e){
            throw e;
        }catch (Throwable t){
            log.error("Unable to execute on event "+request.getEvent(),t);
            String msg = "Unable to execute on event "+request.getEvent()+". Error was "+t.getMessage();
            info.setLastOperationStatus(LifecycleInformation.Status.ERROR);
            info.addErrorMessage(msg);
            report.setStatus(Report.Status.ERROR);
            report.putMessage(msg);
        }
        return report;
    }



    @Override
    public StepExecutionReport performStep(StepExecutionRequest request) throws StepException, InvalidPluginRequestException, InvalidProfileException, ConfigurationException, InsufficientPrivileges {
        StepExecutionReport report = super.performStep(request);
        LifecycleInformation info=report.getToSetLifecycleInformation();
        try {
            if(!canInvokeStep(request.getStep(),request.getCaller(),getConfigurationFromProfile(request.getUseCaseDescriptor())))
                throw new InsufficientPrivileges("Insufficient privileges for executing step "+request.getStep());

            MaterializationPlugin plugin;
            IndexerPluginInterface indexerPlugin;
            plugin= (MaterializationPlugin) pluginManager.getById("SDI-Default-Materializer");
            indexerPlugin = (IndexerPluginInterface) pluginManager.getById("SDI-Indexer-Plugin");



            switch (request.getStep()) {
                case STEPS.SUBMIT:{
                    //TODO validation

                    if(!request.getDocument().getLifecycleInformation().getPhase().equals(LifecycleInformation.DRAFT_PHASE))
                        throw new StepException("Document is not in "+LifecycleInformation.DRAFT_PHASE+" phase");
                    //Materialize layers
                    MaterializationRequest matReq = new MaterializationRequest(request.getUseCaseDescriptor(),request.getCaller(), request.getContext(), request.getDocument());

                    matReq.setDocument(request.getDocument());
                    matReq.setUseCaseDescriptor(request.getUseCaseDescriptor());
                    Document params = new Document();
                    String workspace = request.getUseCaseDescriptor().getId() + request.getContext().getId();
                    params.put("workspace", Files.fixFilename(workspace));


                    matReq.setCallParameters(params);
                    MaterializationReport matRep = plugin.materialize(matReq);


                    report.setResultingDocument(matRep.getResultingDocument());


                    switch(matRep.getStatus()){
                        case OK : {
                            index(report,indexerPlugin);
                            //TODO Optional Notifications
                            break;
                        }
                        case ERROR : {
                            info.setLastOperationStatus(LifecycleInformation.Status.ERROR);
                            matRep.getMessages().forEach(s -> info.addErrorMessage(s));
                            break;
                        }
                        case WARNING : {
                            info.setLastOperationStatus(LifecycleInformation.Status.WARNING);
                            matRep.getMessages().forEach(s -> info.addWarningMessage(s));
                            break;
                        }
                    }

                }
                break;
                case STEPS.REJECT:{
                    if(!request.getDocument().getLifecycleInformation().getPhase().equals(Phases.PENDING_APPROVAL))
                        throw new StepException("Document is not in "+Phases.PENDING_APPROVAL+" phase");
                    // TODO OPTIONAL Notification

                    info.setPhase(Phases.PENDING_APPROVAL);
                    if(request.getCallParameters()!=null&&request.getCallParameters().containsKey(PARAMETERS.NOTES))
                    info.setNotes(request.getCallParameters().getString(PARAMETERS.NOTES));
                    break;
                }

                case STEPS.APPROVE:{
                    // Index-published
                    if(!request.getDocument().getLifecycleInformation().getPhase()
                            .equals(Phases.PENDING_APPROVAL))
                        throw new StepException("Document is not in "+Phases.PENDING_APPROVAL+" phase");

                   index(report,indexerPlugin);
                   break;
                }

                default:
                    throw new StepException("Invalid Step " + request.getStep());
            }
        }catch (StepException e){
            throw e;
        }catch (Throwable t){
            log.error("Unable to perform step "+request.getStep(),t);
            String msg = "Unable to execute Step "+request.getStep()+". Error was "+t.getMessage();
            report.setStatus(Report.Status.ERROR);
            report.putMessage(msg);
            info.setLastOperationStatus(LifecycleInformation.Status.ERROR);
            info.addErrorMessage(msg);
        }
        return report;
    }



    private IndexDocumentReport index(StepExecutionReport report,IndexerPluginInterface indexer) throws InvalidPluginRequestException {
        StepExecutionRequest request = report.getTheRequest();
        LifecycleInformation info=report.getToSetLifecycleInformation();
        IndexDocumentRequest indexRequest = new IndexDocumentRequest(
                request.getUseCaseDescriptor(),request.getCaller(), request.getContext(),request.getDocument());


        indexRequest.setCallParameters(new Document());
        if(request.getStep().equals(STEPS.APPROVE)) {
            // public index
            indexRequest.getCallParameters().putAll(getPublicIndexParams(request));
        }
        else {
            // private index
            indexRequest.getCallParameters().putAll(getPublicIndexParams(request));
        }
        IndexDocumentReport indexReport = indexer.index(indexRequest);


        switch(indexReport.getStatus()){
            case OK : {
                if(request.getStep().equals(STEPS.APPROVE)) {
                    // public index
                    info.setPhase(Phases.PUBLISHED);
                }
                else {
                    // private index
                    info.setPhase(Phases.PENDING_APPROVAL);
                }
                report.setToSetSpatialReference(indexReport.getToSetSpatialReference());
                break;
            }
            case ERROR : {
                info.setLastOperationStatus(LifecycleInformation.Status.ERROR);
                indexReport.getMessages().forEach(s -> info.addErrorMessage(s));
                break;
            }
            case WARNING : {
                info.setLastOperationStatus(LifecycleInformation.Status.WARNING);
                indexReport.getMessages().forEach(s -> info.addWarningMessage(s));
                break;
            }
        }

        return indexReport;

    }




    /**
     * Override this method for programmatic default values management
     *
     * @param currentReport
     * @return
     */
    public EventExecutionReport setDefault(EventExecutionReport currentReport){
        // Default implementation is no op
        return currentReport;
    }

    @Override
    public PluginDescriptor getDescriptor() {
        return DESCRIPTOR;
    }
}
